# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict, List, Optional

from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes


@dataclass
class ToolCall:
    id: Optional[str] = None
    type: Optional[str] = None
    function_name: Optional[str] = None
    function_arguments: Optional[str] = None


@dataclass
class Message:
    role: Optional[str] = None
    content: Optional[str] = None
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


@dataclass
class Choice:
    finish_reason: Optional[str] = None
    role: Optional[str] = None
    content: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


def remove_attributes_with_null_values(attributes: Dict[str, Any]) -> Dict[str, Any]:
    return {attr: value for attr, value in attributes.items() if value is not None}


def attribute_generator(
    original_function: Callable[..., Dict[str, Any]],
) -> Callable[..., Dict[str, Any]]:
    @wraps(original_function)
    def wrapper(*args, **kwargs) -> Dict[str, Any]:
        attributes = original_function(*args, **kwargs)

        return remove_attributes_with_null_values(attributes)

    return wrapper


@attribute_generator
def generate_base_attributes(
    system: GenAIAttributes.GenAiSystemValues,
    operation: GenAIAttributes.GenAiOperationNameValues = GenAIAttributes.GenAiOperationNameValues.CHAT,
) -> Dict[str, Any]:
    attributes = {
        GenAIAttributes.GEN_AI_OPERATION_NAME: operation.value,
        GenAIAttributes.GEN_AI_SYSTEM: system.value,
    }
    return attributes


@attribute_generator
def generate_request_attributes(
    model: Optional[str] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    max_tokens: Optional[int] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
) -> Dict[str, Any]:
    attributes = {
        GenAIAttributes.GEN_AI_REQUEST_MODEL: model,
        GenAIAttributes.GEN_AI_REQUEST_TEMPERATURE: temperature,
        GenAIAttributes.GEN_AI_REQUEST_TOP_P: top_p,
        GenAIAttributes.GEN_AI_REQUEST_MAX_TOKENS: max_tokens,
        GenAIAttributes.GEN_AI_REQUEST_PRESENCE_PENALTY: presence_penalty,
        GenAIAttributes.GEN_AI_REQUEST_FREQUENCY_PENALTY: frequency_penalty,
    }
    return attributes


@attribute_generator
def generate_message_attributes(
    messages: List[Message], capture_content: bool
) -> Dict[str, Any]:
    attributes = {}
    for index, message in enumerate(messages):
        attributes[
            ExtendedGenAIAttributes.GEN_AI_PROMPT_ROLE.format(prompt_index=index)
        ] = message.role

        if capture_content and message.content is not None:
            attributes[
                ExtendedGenAIAttributes.GEN_AI_PROMPT_CONTENT.format(prompt_index=index)
            ] = message.content

        attributes[
            ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALL_ID.format(
                prompt_index=index
            )
        ] = message.tool_call_id
        if message.tool_calls is not None:
            for tool_index, tool_call in enumerate(message.tool_calls):
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_ID.format(
                        prompt_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.id
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_TYPE.format(
                        prompt_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.type
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_FUNCTION_NAME.format(
                        prompt_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.function_name
                if capture_content:
                    attributes[
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                            prompt_index=index, tool_call_index=tool_index
                        )
                    ] = tool_call.function_arguments

    return attributes


@attribute_generator
def generate_response_attributes(
    model: Optional[str] = None,
    finish_reasons: Optional[List[str]] = None,
    id: Optional[str] = None,
    usage_input_tokens: Optional[int] = None,
    usage_output_tokens: Optional[int] = None,
) -> Dict[str, Any]:
    attributes = {
        GenAIAttributes.GEN_AI_RESPONSE_MODEL: model,
        GenAIAttributes.GEN_AI_RESPONSE_FINISH_REASONS: finish_reasons,
        GenAIAttributes.GEN_AI_RESPONSE_ID: id,
        GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS: usage_input_tokens,
        GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS: usage_output_tokens,
    }
    return attributes


@attribute_generator
def generate_choice_attributes(
    choices: List[Choice], capture_content: bool
) -> Dict[str, Any]:
    attributes = {}
    for index, choice in enumerate(choices):
        attributes[
            ExtendedGenAIAttributes.GEN_AI_COMPLETION_FINISH_REASON.format(
                completion_index=index
            )
        ] = choice.finish_reason
        attributes[
            ExtendedGenAIAttributes.GEN_AI_COMPLETION_ROLE.format(
                completion_index=index
            )
        ] = choice.role

        if capture_content and choice.content is not None:
            attributes[
                ExtendedGenAIAttributes.GEN_AI_COMPLETION_CONTENT.format(
                    completion_index=index
                )
            ] = choice.content

        if choice.tool_calls is not None:
            for tool_index, tool_call in enumerate(choice.tool_calls):
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_ID.format(
                        completion_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.id
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_TYPE.format(
                        completion_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.type
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_FUNCTION_NAME.format(
                        completion_index=index, tool_call_index=tool_index
                    )
                ] = tool_call.function_name
                if capture_content:
                    attributes[
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                            completion_index=index, tool_call_index=tool_index
                        )
                    ] = tool_call.function_arguments

    return attributes



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opentelemetry.sdk.trace import ReadableSpan

import llm_tracekit.extended_gen_ai_attributes as ExtendedGenAIAttributes


def assert_messages_in_span(
    span: ReadableSpan, expected_messages: list, expect_content: bool
):
    assert span.attributes is not None

    for index, message in enumerate(expected_messages):
        assert (
            span.attributes[
                ExtendedGenAIAttributes.GEN_AI_PROMPT_ROLE.format(prompt_index=index)
            ]
            == message["role"]
        )

        if "content" in message:
            if expect_content:
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_CONTENT.format(
                            prompt_index=index
                        )
                    ]
                    == message["content"]
                )
            else:
                assert (
                    ExtendedGenAIAttributes.GEN_AI_PROMPT_CONTENT.format(
                        prompt_index=index
                    )
                    not in span.attributes
                )

        if "tool_calls" in message:
            for tool_index, tool_call in enumerate(message["tool_calls"]):
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_ID.format(
                            prompt_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["id"]
                )
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_TYPE.format(
                            prompt_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["type"]
                )
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_FUNCTION_NAME.format(
                            prompt_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["function"]["name"]
                )
                if expect_content:
                    assert (
                        span.attributes[
                            ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                                prompt_index=index, tool_call_index=tool_index
                            )
                        ]
                        == tool_call["function"]["arguments"]
                    )
                else:
                    assert (
                        ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                            prompt_index=index, tool_call_index=tool_index
                        )
                        not in span.attributes
                    )
        else:
            assert (
                ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALLS_ID.format(
                    prompt_index=index, tool_call_index=0
                )
                not in span.attributes
            )

        if "tool_call_id" in message:
            assert (
                span.attributes[
                    ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALL_ID.format(
                        prompt_index=index
                    )
                ]
                == message["tool_call_id"]
            )
        else:
            assert (
                ExtendedGenAIAttributes.GEN_AI_PROMPT_TOOL_CALL_ID.format(
                    prompt_index=index
                )
                not in span.attributes
            )

    # Check that there aren't any additional messages
    assert (
        ExtendedGenAIAttributes.GEN_AI_PROMPT_ROLE.format(prompt_index=index + 1)
        not in span.attributes
    )


def assert_choices_in_span(
    span: ReadableSpan, expected_choices: list, expect_content: bool
):
    assert span.attributes is not None

    for index, choice in enumerate(expected_choices):
        assert (
            span.attributes[
                ExtendedGenAIAttributes.GEN_AI_COMPLETION_ROLE.format(
                    completion_index=index
                )
            ]
            == choice["message"]["role"]
        )
        if "finish_reason" in choice:
            assert (
                span.attributes[
                    ExtendedGenAIAttributes.GEN_AI_COMPLETION_FINISH_REASON.format(
                        completion_index=index
                    )
                ]
                == choice["finish_reason"]
            )
        if "content" in choice["message"]:
            if expect_content:
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_CONTENT.format(
                            completion_index=index
                        )
                    ]
                    == choice["message"]["content"]
                )
            else:
                assert (
                    ExtendedGenAIAttributes.GEN_AI_COMPLETION_CONTENT.format(
                        completion_index=index
                    )
                    not in span.attributes
                )

        if "tool_calls" in choice["message"]:
            for tool_index, tool_call in enumerate(choice["message"]["tool_calls"]):
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_ID.format(
                            completion_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["id"]
                )
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_TYPE.format(
                            completion_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["type"]
                )
                assert (
                    span.attributes[
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_FUNCTION_NAME.format(
                            completion_index=index, tool_call_index=tool_index
                        )
                    ]
                    == tool_call["function"]["name"]
                )
                if expect_content:
                    assert (
                        span.attributes[
                            ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                                completion_index=index, tool_call_index=tool_index
                            )
                        ]
                        == tool_call["function"]["arguments"]
                    )
                else:
                    assert (
                        ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_FUNCTION_ARGUMENTS.format(
                            completion_index=index, tool_call_index=tool_index
                        )
                        not in span.attributes
                    )
        else:
            assert (
                ExtendedGenAIAttributes.GEN_AI_COMPLETION_TOOL_CALLS_ID.format(
                    completion_index=index, tool_call_index=0
                )
                not in span.attributes
            )

    # Check that there aren't any additional choices
    assert (
        ExtendedGenAIAttributes.GEN_AI_COMPLETION_ROLE.format(
            completion_index=index + 1
        )
        not in span.attributes
    )



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import os

import boto3
import pytest
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.sampling import ALWAYS_OFF

from llm_tracekit.bedrock.instrumentor import BedrockInstrumentor
from llm_tracekit.instrumentation_utils import (
    OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT,
)

EVENT_STREAM_CT = "application/vnd.amazon.eventstream"


pytest.register_assert_rewrite("tests.bedrock.utils")


@pytest.fixture(autouse=True)
def bedrock_env_vars():
    for env_var_name in [
        "AWS_ACCESS_KEY_ID",
        "AWS_SECRET_ACCESS_KEY",
        "AWS_SESSION_TOKEN",
    ]:
        if not os.getenv(env_var_name):
            os.environ[env_var_name] = f"test_{env_var_name.lower()}"

    if not os.getenv("AWS_DEFAULT_REGION"):
        os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


@pytest.fixture
def bedrock_client_with_content(instrument_with_content):
    return boto3.client("bedrock-runtime")


@pytest.fixture
def bedrock_client_no_content(instrument_no_content):
    return boto3.client("bedrock-runtime")


@pytest.fixture
def bedrock_agent_client_with_content(instrument_with_content):
    return boto3.client("bedrock-agent-runtime")


@pytest.fixture
def bedrock_agent_client_no_content(instrument_no_content):
    return boto3.client("bedrock-agent-runtime")


@pytest.fixture
def claude_model_id() -> str:
    return "anthropic.claude-3-5-sonnet-20240620-v1:0"


@pytest.fixture
def llama_model_id() -> str:
    return "meta.llama3-8b-instruct-v1:0"


@pytest.fixture
def agent_id() -> str:
    return os.environ.get("AWS_BEDROCK_AGENT_ID", "test_agent_id")


@pytest.fixture
def agent_alias_id() -> str:
    return os.environ.get("AWS_BEDROCK_AGENT_ALIAS_ID", "test_agent_alias_id")


def handle_recording_boto_response(response: dict) -> dict:
    """Prepares boto3 reponses for recoring/playback.

    Args:
        response: HTTP response

    Returns:
        Fixed response

    Notes:
        * This function is required to play back streaming and invoke_model output
        * This function was generated by ChatGPT
    """
    # runs both when recording and when loading the cassette
    headers = response["headers"]
    if headers.get("x-vcr-base64") == ["yes"]:
        # playback branch ─ decode back to raw bytes
        response["body"]["string"] = base64.b64decode(response["body"]["string"])
        headers.pop("x-vcr-base64")
        return response

    if EVENT_STREAM_CT in (headers.get("Content-Type") or [""])[0]:
        # record branch ─ protect the bytes from PyYAML
        raw = response["body"]["string"]  # bytes from socket
        response["body"]["string"] = base64.b64encode(raw).decode()
        headers["x-vcr-base64"] = ["yes"]

    # Fix Content-Length header
    response["headers"]["Content-Length"] = [str(len(response["body"]["string"]))]
    return response


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": [
            ("X-Amz-Security-Token", "test-security-token"),
            ("Authorization", "test-auth"),
        ],
        "decode_compressed_response": True,
        "before_record_response": handle_recording_boto_response,
    }


@pytest.fixture(scope="function")
def instrument_no_content(tracer_provider, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "False"})

    instrumentor = BedrockInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()


@pytest.fixture(scope="function")
def instrument_with_content(tracer_provider, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "True"})
    instrumentor = BedrockInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()


@pytest.fixture(scope="function")
def instrument_with_content_unsampled(span_exporter, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "True"})

    tracer_provider = TracerProvider(sampler=ALWAYS_OFF)
    tracer_provider.add_span_processor(SimpleSpanProcessor(span_exporter))

    instrumentor = BedrockInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()



import pytest
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv._incubating.attributes import (
    server_attributes as ServerAttributes,
)
from opentelemetry.semconv._incubating.metrics import gen_ai_metrics

_DURATION_BUCKETS = (
    0.01,
    0.02,
    0.04,
    0.08,
    0.16,
    0.32,
    0.64,
    1.28,
    2.56,
    5.12,
    10.24,
    20.48,
    40.96,
    81.92,
)
_TOKEN_USAGE_BUCKETS = (
    1,
    4,
    16,
    64,
    256,
    1024,
    4096,
    16384,
    65536,
    262144,
    1048576,
    4194304,
    16777216,
    67108864,
)


def assert_all_metric_attributes(data_point):
    assert GenAIAttributes.GEN_AI_OPERATION_NAME in data_point.attributes
    assert (
        data_point.attributes[GenAIAttributes.GEN_AI_OPERATION_NAME]
        == GenAIAttributes.GenAiOperationNameValues.CHAT.value
    )
    assert GenAIAttributes.GEN_AI_SYSTEM in data_point.attributes
    assert (
        data_point.attributes[GenAIAttributes.GEN_AI_SYSTEM]
        == GenAIAttributes.GenAiSystemValues.OPENAI.value
    )
    assert GenAIAttributes.GEN_AI_REQUEST_MODEL in data_point.attributes
    assert data_point.attributes[GenAIAttributes.GEN_AI_REQUEST_MODEL] == "gpt-4o-mini"
    assert GenAIAttributes.GEN_AI_RESPONSE_MODEL in data_point.attributes
    assert (
        data_point.attributes[GenAIAttributes.GEN_AI_RESPONSE_MODEL]
        == "gpt-4o-mini-2024-07-18"
    )
    assert (
        GenAIAttributes.GEN_AI_OPENAI_RESPONSE_SYSTEM_FINGERPRINT
        in data_point.attributes
    )
    assert (
        data_point.attributes[GenAIAttributes.GEN_AI_OPENAI_RESPONSE_SYSTEM_FINGERPRINT]
        == "fp_0ba0d124f1"
    )
    assert GenAIAttributes.GEN_AI_OPENAI_RESPONSE_SERVICE_TIER in data_point.attributes
    assert (
        data_point.attributes[GenAIAttributes.GEN_AI_OPENAI_RESPONSE_SERVICE_TIER]
        == "default"
    )
    assert data_point.attributes[ServerAttributes.SERVER_ADDRESS] == "api.openai.com"


@pytest.mark.vcr()
def test_chat_completion_metrics(metric_reader, openai_client, instrument_with_content):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert len(metric_data) == 2

    duration_metric = next(
        (
            m
            for m in metric_data
            if m.name == gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION
        ),
        None,
    )
    assert duration_metric is not None

    duration_point = duration_metric.data.data_points[0]
    assert duration_point.sum > 0
    assert_all_metric_attributes(duration_point)
    assert duration_point.explicit_bounds == _DURATION_BUCKETS

    token_usage_metric = next(
        (m for m in metric_data if m.name == gen_ai_metrics.GEN_AI_CLIENT_TOKEN_USAGE),
        None,
    )
    assert token_usage_metric is not None

    input_token_usage = next(
        (
            d
            for d in token_usage_metric.data.data_points
            if d.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
            == GenAIAttributes.GenAiTokenTypeValues.INPUT.value
        ),
        None,
    )
    assert input_token_usage is not None
    assert input_token_usage.sum == 12

    assert input_token_usage.explicit_bounds == _TOKEN_USAGE_BUCKETS
    assert input_token_usage.bucket_counts[2] == 1
    assert_all_metric_attributes(input_token_usage)

    output_token_usage = next(
        (
            d
            for d in token_usage_metric.data.data_points
            if d.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
            == GenAIAttributes.GenAiTokenTypeValues.COMPLETION.value
        ),
        None,
    )
    assert output_token_usage is not None
    assert output_token_usage.sum == 5
    # assert against buckets [1, 4, 16, 64, 256, 1024, 4096, 16384, 65536, 262144, 1048576, 4194304, 16777216, 67108864]
    assert output_token_usage.bucket_counts[2] == 1
    assert_all_metric_attributes(output_token_usage)


@pytest.mark.vcr()
@pytest.mark.asyncio()
async def test_async_chat_completion_metrics(
    metric_reader, async_openai_client, instrument_with_content
):
    llm_model_value = "gpt-4o-mini"
    messages_value = [{"role": "user", "content": "Say this is a test"}]

    await async_openai_client.chat.completions.create(
        messages=messages_value, model=llm_model_value, stream=False
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert len(metric_data) == 2

    duration_metric = next(
        (
            m
            for m in metric_data
            if m.name == gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION
        ),
        None,
    )
    assert duration_metric is not None
    assert duration_metric.data.data_points[0].sum > 0
    assert_all_metric_attributes(duration_metric.data.data_points[0])

    token_usage_metric = next(
        (m for m in metric_data if m.name == gen_ai_metrics.GEN_AI_CLIENT_TOKEN_USAGE),
        None,
    )
    assert token_usage_metric is not None

    input_token_usage = next(
        (
            d
            for d in token_usage_metric.data.data_points
            if d.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
            == GenAIAttributes.GenAiTokenTypeValues.INPUT.value
        ),
        None,
    )

    assert input_token_usage is not None
    assert input_token_usage.sum == 12
    assert_all_metric_attributes(input_token_usage)

    output_token_usage = next(
        (
            d
            for d in token_usage_metric.data.data_points
            if d.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
            == GenAIAttributes.GenAiTokenTypeValues.COMPLETION.value
        ),
        None,
    )

    assert output_token_usage is not None
    assert output_token_usage.sum == 12
    assert_all_metric_attributes(output_token_usage)



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest
from openai import AsyncOpenAI, OpenAI
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.sampling import ALWAYS_OFF

from llm_tracekit.instrumentation_utils import (
    OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT,
)
from llm_tracekit.openai.instrumentor import OpenAIInstrumentor


@pytest.fixture(autouse=True)
def openai_env_vars():
    if not os.getenv("OPENAI_API_KEY"):
        os.environ["OPENAI_API_KEY"] = "test_openai_api_key"


@pytest.fixture
def openai_client():
    return OpenAI()


@pytest.fixture
def async_openai_client():
    return AsyncOpenAI()


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": [
            ("cookie", "test_cookie"),
            ("authorization", "Bearer test_openai_api_key"),
            ("openai-organization", "test_openai_org_id"),
            ("openai-project", "test_openai_project_id"),
        ],
        "decode_compressed_response": True,
        "before_record_response": scrub_response_headers,
    }


@pytest.fixture(scope="function")
def instrument_no_content(tracer_provider, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "False"})

    instrumentor = OpenAIInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()


@pytest.fixture(scope="function")
def instrument_with_content(tracer_provider, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "True"})
    instrumentor = OpenAIInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()


@pytest.fixture(scope="function")
def instrument_with_content_unsampled(span_exporter, meter_provider):
    os.environ.update({OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT: "True"})

    tracer_provider = TracerProvider(sampler=ALWAYS_OFF)
    tracer_provider.add_span_processor(SimpleSpanProcessor(span_exporter))

    instrumentor = OpenAIInstrumentor()
    instrumentor.instrument(
        tracer_provider=tracer_provider,
        meter_provider=meter_provider,
    )

    yield instrumentor
    os.environ.pop(OTEL_INSTRUMENTATION_GENAI_CAPTURE_MESSAGE_CONTENT, None)
    instrumentor.uninstrument()


def scrub_response_headers(response):
    """
    This scrubs sensitive response headers. Note they are case-sensitive!
    """
    response["headers"]["openai-organization"] = "test_openai_org_id"
    response["headers"]["Set-Cookie"] = "test_set_cookie"
    return response



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Generator

import pytest
import yaml
from opentelemetry.sdk.metrics import (
    MeterProvider,
)
from opentelemetry.sdk.metrics.export import (
    InMemoryMetricReader,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)

pytest.register_assert_rewrite("tests.utils")


@pytest.fixture(scope="function", name="span_exporter")
def fixture_span_exporter() -> Generator[InMemorySpanExporter, None, None]:
    exporter = InMemorySpanExporter()
    yield exporter


@pytest.fixture(scope="function", name="metric_reader")
def fixture_metric_reader() -> Generator[InMemoryMetricReader, None, None]:
    exporter = InMemoryMetricReader()
    yield exporter


@pytest.fixture(scope="function", name="tracer_provider")
def fixture_tracer_provider(span_exporter) -> TracerProvider:
    provider = TracerProvider()
    provider.add_span_processor(SimpleSpanProcessor(span_exporter))
    return provider


@pytest.fixture(scope="function", name="meter_provider")
def fixture_meter_provider(metric_reader) -> MeterProvider:
    meter_provider = MeterProvider(
        metric_readers=[metric_reader],
    )

    return meter_provider


class LiteralBlockScalar(str):
    """Formats the string as a literal block scalar, preserving whitespace and
    without interpreting escape characters"""


def literal_block_scalar_presenter(dumper, data):
    """Represents a scalar string as a literal block, via '|' syntax"""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


yaml.add_representer(LiteralBlockScalar, literal_block_scalar_presenter)


def process_string_value(string_value):
    """Pretty-prints JSON or returns long strings as a LiteralBlockScalar"""
    try:
        json_data = json.loads(string_value)
        return LiteralBlockScalar(json.dumps(json_data, indent=2))
    except (ValueError, TypeError):
        if len(string_value) > 80:
            return LiteralBlockScalar(string_value)
    return string_value


def convert_body_to_literal(data):
    """Searches the data for body strings, attempting to pretty-print JSON"""
    if isinstance(data, dict):
        for key, value in data.items():
            # Handle response body case (e.g., response.body.string)
            if key == "body" and isinstance(value, dict) and "string" in value:
                value["string"] = process_string_value(value["string"])

            # Handle request body case (e.g., request.body)
            elif key == "body" and isinstance(value, str):
                data[key] = process_string_value(value)

            else:
                convert_body_to_literal(value)

    elif isinstance(data, list):
        for idx, choice in enumerate(data):
            data[idx] = convert_body_to_literal(choice)

    return data


class PrettyPrintJSONBody:
    """This makes request and response body recordings more readable."""

    @staticmethod
    def serialize(cassette_dict):
        cassette_dict = convert_body_to_literal(cassette_dict)
        return yaml.dump(cassette_dict, default_flow_style=False, allow_unicode=True)

    @staticmethod
    def deserialize(cassette_string):
        return yaml.load(cassette_string, Loader=yaml.Loader)


@pytest.fixture(scope="module", autouse=True)
def fixture_vcr(vcr):
    vcr.register_serializer("yaml", PrettyPrintJSONBody)
    return vcr



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import suppress

from llm_tracekit.coralogix import (
    setup_export_to_coralogix as setup_export_to_coralogix,
)

with suppress(ImportError):
    from llm_tracekit.openai.instrumentor import (
        OpenAIInstrumentor as OpenAIInstrumentor,
    )

with suppress(ImportError):
    from llm_tracekit.bedrock.instrumentor import (
        BedrockInstrumentor as BedrockInstrumentor,
    )



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Collection

import botocore.client
import botocore.session
from opentelemetry.instrumentation.instrumentor import (  # type: ignore[attr-defined] # Mypy doesn't recognize the attribute
    BaseInstrumentor,
)
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.metrics import get_meter
from opentelemetry.semconv.schemas import Schemas
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from llm_tracekit.bedrock.package import _instruments
from llm_tracekit.bedrock.patch import create_client_wrapper
from llm_tracekit.instrumentation_utils import is_content_enabled
from llm_tracekit.instruments import Instruments


class BedrockInstrumentor(BaseInstrumentor):
    def __init__(self):
        self._meter = None

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        """Enable OpenAI instrumentation."""
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(
            __name__,
            "",
            tracer_provider,
            schema_url=Schemas.V1_32_0.value,
        )
        meter_provider = kwargs.get("meter_provider")
        self._meter = get_meter(
            __name__,
            "",
            meter_provider,
            schema_url=Schemas.V1_32_0.value,
        )

        instruments = Instruments(self._meter)

        wrap_function_wrapper(
            module="botocore.client",
            name="ClientCreator.create_client",
            wrapper=create_client_wrapper(tracer, instruments, is_content_enabled()),
        )

        wrap_function_wrapper(
            module="botocore.session",
            name="Session.create_client",
            wrapper=create_client_wrapper(tracer, instruments, is_content_enabled()),
        )

    def _uninstrument(self, **kwargs):
        unwrap(botocore.client.ClientCreator, "create_client")
        unwrap(botocore.session.Session, "create_client")



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Optional

from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv.attributes import error_attributes as ErrorAttributes

from llm_tracekit.instruments import Instruments


def record_metrics(
    instruments: Instruments,
    duration: float,
    request_model: Optional[str] = None,
    response_model: Optional[str] = None,
    usage_input_tokens: Optional[int] = None,
    usage_output_tokens: Optional[int] = None,
    error_type: Optional[str] = None,
):
    common_attributes = {
        GenAIAttributes.GEN_AI_OPERATION_NAME: GenAIAttributes.GenAiOperationNameValues.CHAT.value,
        GenAIAttributes.GEN_AI_SYSTEM: GenAIAttributes.GenAiSystemValues.AWS_BEDROCK.value,
    }

    if request_model is not None:
        common_attributes[GenAIAttributes.GEN_AI_REQUEST_MODEL] = request_model
    if response_model is not None:
        common_attributes[GenAIAttributes.GEN_AI_RESPONSE_MODEL] = response_model

    if error_type:
        common_attributes[ErrorAttributes.ERROR_TYPE] = error_type

    instruments.operation_duration_histogram.record(
        duration,
        attributes=common_attributes,
    )

    if usage_input_tokens is not None:
        input_attributes = {
            **common_attributes,
            GenAIAttributes.GEN_AI_TOKEN_TYPE: GenAIAttributes.GenAiTokenTypeValues.INPUT.value,
        }
        instruments.token_usage_histogram.record(
            usage_input_tokens,
            attributes=input_attributes,
        )

    if usage_output_tokens is not None:
        completion_attributes = {
            **common_attributes,
            GenAIAttributes.GEN_AI_TOKEN_TYPE: GenAIAttributes.GenAiTokenTypeValues.COMPLETION.value,
        }
        instruments.token_usage_histogram.record(
            usage_output_tokens,
            attributes=completion_attributes,
        )


def decode_tool_use_in_stream(tool_use):
    # input get sent encoded in json
    if "input" in tool_use:
        try:
            tool_use["input"] = json.loads(tool_use["input"])
        except json.JSONDecodeError:
            pass
    return tool_use



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from timeit import default_timer
from typing import Any, Callable, Dict, Optional

from botocore.eventstream import EventStream, EventStreamError
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.trace import Span
from wrapt import ObjectProxy

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes
from llm_tracekit.bedrock.utils import record_metrics
from llm_tracekit.instruments import Instruments
from llm_tracekit.span_builder import (
    Choice,
    Message,
    attribute_generator,
    generate_base_attributes,
    generate_choice_attributes,
    generate_message_attributes,
    generate_response_attributes,
)


@attribute_generator
def generate_attributes_from_invoke_agent_input(
    kwargs: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    base_attributes = generate_base_attributes(
        system=GenAIAttributes.GenAiSystemValues.AWS_BEDROCK
    )
    message_attributes = generate_message_attributes(
        messages=[Message(role="user", content=kwargs.get("inputText"))],
        capture_content=capture_content,
    )

    attributes = {
        **base_attributes,
        **message_attributes,
        GenAIAttributes.GEN_AI_AGENT_ID: kwargs.get("agentId"),
        ExtendedGenAIAttributes.GEN_AI_BEDROCK_AGENT_ALIAS_ID: kwargs.get(
            "agentAliasId"
        ),
    }

    return attributes


def record_invoke_agent_result_attributes(
    content: Optional[str],
    usage_input_tokens: Optional[int],
    usage_output_tokens: Optional[int],
    span: Span,
    start_time: float,
    instruments: Instruments,
    capture_content: bool,
):
    try:
        attributes = {
            **generate_response_attributes(
                usage_input_tokens=usage_input_tokens,
                usage_output_tokens=usage_output_tokens,
            ),
            **generate_choice_attributes(
                choices=[Choice(role="assistant", content=content)],
                capture_content=capture_content,
            ),
        }
        span.set_attributes(attributes)
    finally:
        duration = max((default_timer() - start_time), 0)
        span.end()
        record_metrics(
            instruments=instruments,
            duration=duration,
            usage_input_tokens=usage_input_tokens,
            usage_output_tokens=usage_output_tokens,
        )


class InvokeAgentStreamWrapper(ObjectProxy):
    """Wrapper for botocore.eventstream.EventStream"""

    def __init__(
        self,
        stream: EventStream,
        stream_done_callback: Callable[
            [Optional[str], Optional[int], Optional[int]], None
        ],
        stream_error_callback: Callable[[Exception], None],
    ):
        super().__init__(stream)

        self._stream_done_callback = stream_done_callback
        self._stream_error_callback = stream_error_callback
        self._content = None
        self._usage_input_tokens: Optional[int] = None
        self._usage_output_tokens: Optional[int] = None

    def __iter__(self):
        try:
            for event in self.__wrapped__:
                self._process_event(event)
                yield event

            self._stream_done_callback(
                self._content, self._usage_input_tokens, self._usage_output_tokens
            )
        except EventStreamError as exc:
            self._stream_error_callback(exc)
            raise

    def _process_usage_data(self, usage: Dict[str, int]):
        if "inputTokens" in usage:
            if self._usage_input_tokens is None:
                self._usage_input_tokens = 0

            self._usage_input_tokens += usage["inputTokens"]

        if "outputTokens" in usage:
            if self._usage_output_tokens is None:
                self._usage_output_tokens = 0

            self._usage_output_tokens += usage["outputTokens"]

    def _process_event(self, event):
        if "chunk" in event:
            if self._content is None:
                self._content = ""

            encoded_content = event["chunk"].get("bytes")
            if encoded_content is not None:
                self._content += encoded_content.decode()

        if "trace" in event and "trace" in event["trace"]:
            for key in [
                "preProcessingTrace",
                "postProcessingTrace",
                "orchestrationTrace",
                "routingClassifierTrace",
            ]:
                usage_data = (
                    event["trace"]["trace"]
                    .get(key, {})
                    .get("modelInvocationOutput", {})
                    .get("metadata", {})
                    .get("usage")
                )
                if usage_data is not None:
                    self._process_usage_data(usage_data)



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import pytest
from botocore.exceptions import ClientError

from tests.bedrock.utils import assert_attributes_in_span, assert_expected_metrics
from tests.utils import assert_choices_in_span, assert_messages_in_span


def _run_and_check_invoke_agent(
    bedrock_agent_client,
    span_exporter,
    metric_reader,
    agent_id: str,
    agent_alias_id: str,
    expect_content: bool,
    session_id: str,
):
    result = bedrock_agent_client.invoke_agent(
        agentAliasId=agent_alias_id,
        agentId=agent_id,
        inputText="say this is a test",
        sessionId=session_id,
        enableTrace=True,
    )
    response_text = ""
    usage_input_tokens = 0
    usage_output_tokens = 0
    for event in result["completion"]:
        if "chunk" in event:
            response_text += event["chunk"]["bytes"].decode()
        if "trace" in event:
            for key in [
                "preProcessingTrace",
                "postProcessingTrace",
                "orchestrationTrace",
                "routingClassifierTrace",
            ]:
                usage_data = (
                    event["trace"]["trace"]
                    .get(key, {})
                    .get("modelInvocationOutput", {})
                    .get("metadata", {})
                    .get("usage", {})
                )
                usage_input_tokens += usage_data.get("inputTokens", 0)
                usage_output_tokens += usage_data.get("outputTokens", 0)

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_agent",
        agent_id=agent_id,
        agent_alias_id=agent_alias_id,
        usage_input_tokens=usage_input_tokens,
        usage_output_tokens=usage_output_tokens,
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages,
        expect_content=expect_content,
    )

    expected_choice = {
        "message": {
            "role": "assistant",
            "content": response_text,
        }
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[expected_choice], expect_content=expect_content
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        usage_input_tokens=usage_input_tokens,
        usage_output_tokens=usage_output_tokens,
    )


@pytest.mark.vcr()
def test_invoke_agent_with_content(
    bedrock_agent_client_with_content,
    span_exporter,
    metric_reader,
    agent_id: str,
    agent_alias_id: str,
):
    _run_and_check_invoke_agent(
        bedrock_agent_client=bedrock_agent_client_with_content,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        agent_id=agent_id,
        agent_alias_id=agent_alias_id,
        session_id="11",
        expect_content=True,
    )


@pytest.mark.vcr()
def test_invoke_agent_no_content(
    bedrock_agent_client_no_content,
    span_exporter,
    metric_reader,
    agent_id: str,
    agent_alias_id: str,
):
    _run_and_check_invoke_agent(
        bedrock_agent_client=bedrock_agent_client_no_content,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        agent_id=agent_id,
        agent_alias_id=agent_alias_id,
        session_id="22",
        expect_content=False,
    )


@pytest.mark.vcr()
def test_invoke_agent_bad_auth(
    instrument_with_content,
    span_exporter,
    metric_reader,
    agent_id: str,
    agent_alias_id: str,
):
    client = boto3.client(
        "bedrock-agent-runtime",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        aws_session_token="test",
    )
    with pytest.raises(ClientError):
        client.invoke_agent(
            agentAliasId=agent_alias_id,
            agentId=agent_id,
            inputText="say this is a test",
            sessionId="123456",
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_agent",
        agent_id=agent_id,
        agent_alias_id=agent_alias_id,
        error="ClientError",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        error="ClientError",
    )


@pytest.mark.vcr()
def test_invoke_agent_non_existing_agent(
    bedrock_agent_client_with_content, span_exporter, metric_reader
):
    agent_id = "agent_id"
    agent_alias_id = "agent_alias"
    with pytest.raises(Exception):
        bedrock_agent_client_with_content.invoke_agent(
            agentAliasId=agent_alias_id,
            agentId=agent_id,
            inputText="say this is a test",
            sessionId="123456",
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_agent",
        agent_id=agent_id,
        agent_alias_id=agent_alias_id,
        error="ValidationException",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        error="ValidationException",
    )



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor

from llm_tracekit.instrumentation_utils import enable_capture_content


def setup_export_to_coralogix(
    service_name: str,
    coralogix_token: Optional[str] = None,
    coralogix_endpoint: Optional[str] = None,
    application_name: Optional[str] = None,
    subsystem_name: Optional[str] = None,
    use_batch_processor: bool = False,
    capture_content: bool = True,
):
    """
    Setup OpenAI spans to be exported to Coralogix.

    Args:
        service_name: The service name.
        coralogix_token: The Coralogix token. Defaults to os.environ["CX_TOKEN]
        coralogix_endpoint: The Coralogix endpoint. Defaults to os.environ["CX_ENDPOINT"]
        application_name: The Coralogix application name. Defaults to os.environ["CX_APPLICATION_NAME"]
        subsystem_name: The Coralogix subsystem name. Defaults to os.environ["CX_SUBSYSTEM_NAME"]
        use_batch_processor: Whether to use a batch processor or a simple processor..
        capture_content: Whether to capture the content of the messages.
    """

    # Read environment variables as defaults if needed
    if coralogix_token is None:
        coralogix_token = os.environ["CX_TOKEN"]
    if coralogix_endpoint is None:
        coralogix_endpoint = os.environ["CX_ENDPOINT"]
    if application_name is None:
        application_name = os.environ["CX_APPLICATION_NAME"]
    if subsystem_name is None:
        subsystem_name = os.environ["CX_SUBSYSTEM_NAME"]
    if capture_content:
        enable_capture_content()

    # set up a tracer provider to send spans to coralogix.
    tracer_provider = TracerProvider(
        resource=Resource.create({SERVICE_NAME: service_name}),
    )

    # set up an OTLP exporter to send spans to coralogix directly.
    headers = {
        "authorization": f"Bearer {coralogix_token}",
        "cx-application-name": application_name,
        "cx-subsystem-name": subsystem_name,
    }
    exporter = OTLPSpanExporter(endpoint=coralogix_endpoint, headers=headers)

    # set up a span processor to send spans to the exporter
    span_processor = (
        BatchSpanProcessor(exporter)
        if use_batch_processor
        else SimpleSpanProcessor(exporter)
    )

    # add the span processor to the tracer provider
    tracer_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(tracer_provider)



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from contextlib import suppress
from copy import deepcopy
from timeit import default_timer
from typing import Any, Callable, Dict, List, Optional, Union

from botocore.eventstream import EventStream, EventStreamError
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.trace import Span
from wrapt import ObjectProxy

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes
from llm_tracekit.bedrock.utils import decode_tool_use_in_stream, record_metrics
from llm_tracekit.instruments import Instruments
from llm_tracekit.span_builder import (
    Choice,
    Message,
    ToolCall,
    attribute_generator,
    generate_base_attributes,
    generate_choice_attributes,
    generate_message_attributes,
    generate_request_attributes,
    generate_response_attributes,
)


def _combine_tool_call_content_blocks(
    content_blocks: List[Dict[str, Any]],
) -> Optional[str]:
    text_blocks = []
    for content_block in content_blocks:
        if "text" in content_block:
            text_blocks.append(content_block["text"])

        if "json" in content_block:
            return json.dumps(content_block["json"])

    if len(text_blocks) > 0:
        return "".join(text_blocks)

    return None


def _parse_converse_message(
    role: Optional[str], content_blocks: Optional[List[Dict[str, Any]]]
) -> List[Message]:
    """Attempts to combine the content blocks of a `converse` API message to a single message."""
    if content_blocks is None:
        return [Message(role=role)]

    text_blocks = []
    tool_call_blocks = []
    tool_call_result_blocks = []

    # Get all the content blocks we support
    for content_block in content_blocks:
        if "text" in content_block:
            text_blocks.append(content_block["text"])

        if "toolUse" in content_block:
            tool_call_blocks.append(content_block["toolUse"])

        if "toolResult" in content_block:
            tool_call_result_blocks.append(content_block["toolResult"])

    # We follow the same logic the OTEL implementation uses:
    #  * For assistant messages (text/tool calls) - return a single message
    #  * for user messages (text / tool call results) - return a message for each tool 
    #    call result, and another message for text
    messages = []
    if role == "assistant":
        tool_calls = None
        content = None
        if len(text_blocks) > 0:
            content="".join(text_blocks)
        if len(tool_call_blocks) > 0:
            tool_calls = []
            for tool_call in tool_call_blocks:
                arguments = None
                if "input" in tool_call:
                    arguments = json.dumps(tool_call["input"])

                tool_calls.append(
                    ToolCall(
                        id=tool_call.get("toolUseId"),
                        type="function",
                        function_name=tool_call.get("name"),
                        function_arguments=arguments,
                    )
                )

        messages.append(
            Message(
                role=role,
                tool_calls=tool_calls,
                content=content
            )
        )
    elif role == "user":
        if len(tool_call_result_blocks) > 0:
            for tool_call_result in tool_call_result_blocks:
                content = None
                if "content" in tool_call_result:
                    content = _combine_tool_call_content_blocks(tool_call_result["content"])

                messages.append(
                    Message(
                        role="tool",
                        tool_call_id=tool_call_result.get("toolUseId"),
                        content=content,
                    )
                )

        if len(text_blocks) > 0:
            messages.append(Message(role=role, content="".join(text_blocks)))

    if len(messages) > 0:
        return messages

    return [Message(role=role)]


@attribute_generator
def generate_attributes_from_converse_input(
    kwargs: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    inference_config = kwargs.get("inferenceConfig", {})
    messages = []
    for system_message in kwargs.get("system", []):
        messages.append(Message(role="system", content=system_message.get("text")))

    for message in kwargs.get("messages", []):
        messages.extend(
            _parse_converse_message(
                role=message.get("role"), content_blocks=message.get("content")
            )
        )

    tool_attributes = {}
    tool_configs = kwargs.get("toolConfig", {}).get("tools", [])
    # tool configs can contain either "toolSpec" (which is the actual tool definition) or
    # "cachePoint" (to use prompt caching) - we can only gather information from "toolSpec",
    # so we filter out the rest
    tool_specs = [tool["toolSpec"] for tool in tool_configs if "toolSpec" in tool]
    for index, tool_spec in enumerate(tool_specs):
        tool_params = None
        if "inputSchema" in tool_spec and "json" in tool_spec["inputSchema"]:
            with suppress(TypeError):
                tool_params = json.dumps(tool_spec["inputSchema"]["json"])

        tool_attributes.update(
            {
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_NAME.format(
                    tool_index=index
                ): tool_spec.get("name"),
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_DESCRIPTION.format(
                    tool_index=index
                ): tool_spec.get("description"),
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_PARAMETERS.format(
                    tool_index=index
                ): tool_params,
            }
        )

    return {
        **generate_base_attributes(
            system=GenAIAttributes.GenAiSystemValues.AWS_BEDROCK
        ),
        **generate_request_attributes(
            model=kwargs.get("modelId"),
            temperature=inference_config.get("temperature"),
            top_p=inference_config.get("topP"),
            max_tokens=inference_config.get("maxTokens"),
        ),
        **generate_message_attributes(
            messages=messages, capture_content=capture_content
        ),
        **tool_attributes,
    }


def record_converse_result_attributes(
    result: Dict[str, Any],
    span: Span,
    start_time: float,
    instruments: Instruments,
    capture_content: bool,
    model: Optional[str],
):
    finish_reason = result.get("stopReason")
    usage_data = result.get("usage", {})
    usage_input_tokens = usage_data.get("inputTokens")
    usage_output_tokens = usage_data.get("outputTokens")

    response_attributes = generate_response_attributes(
        model=model,
        finish_reasons=None if finish_reason is None else [finish_reason],
        usage_input_tokens=usage_input_tokens,
        usage_output_tokens=usage_output_tokens,
    )
    span.set_attributes(response_attributes)

    response_message = result.get("output", {}).get("message")
    if response_message is not None:
        parsed_response_message = _parse_converse_message(
            role=response_message.get("role"),
            content_blocks=response_message.get("content"),
        )[0]
        choice = Choice(
            finish_reason=finish_reason,
            role=parsed_response_message.role,
            content=parsed_response_message.content,
            tool_calls=parsed_response_message.tool_calls,
        )
        span.set_attributes(
            generate_choice_attributes(
                choices=[choice], capture_content=capture_content
            )
        )

    span.end()

    duration = max((default_timer() - start_time), 0)
    record_metrics(
        instruments=instruments,
        duration=duration,
        request_model=model,
        response_model=model,
        usage_input_tokens=usage_input_tokens,
        usage_output_tokens=usage_output_tokens,
    )


class ConverseStreamWrapper(ObjectProxy):
    """Wrapper for botocore.eventstream.EventStream"""

    def __init__(
        self,
        stream: EventStream,
        stream_done_callback: Callable[[Dict[str, Union[int, str]]], None],
        stream_error_callback: Callable[[Exception], None],
    ):
        super().__init__(stream)

        self._stream_done_callback = stream_done_callback
        self._stream_error_callback = stream_error_callback
        # accumulating things in the same shape of non-streaming version
        # {"usage": {"inputTokens": 0, "outputTokens": 0}, "stopReason": "finish", "output": {"message": {"role": "", "content": [{"text": ""}]}
        self._response: Dict[str, Any] = {}
        self._message = None
        self._content_block: Dict[str, Any] = {}
        self._record_message = False

    def __iter__(self):
        try:
            for event in self.__wrapped__:
                self._process_event(event)
                yield event
        except EventStreamError as exc:
            self._stream_error_callback(exc)
            raise

    def _process_event(self, event):
        # pylint: disable=too-many-branches
        if "messageStart" in event:
            # {'messageStart': {'role': 'assistant'}}
            if event["messageStart"].get("role") == "assistant":
                self._record_message = True
                self._message = {"role": "assistant", "content": []}
            return

        if "contentBlockStart" in event:
            # {'contentBlockStart': {'start': {'toolUse': {'toolUseId': 'id', 'name': 'func_name'}}, 'contentBlockIndex': 1}}
            start = event["contentBlockStart"].get("start", {})
            if "toolUse" in start:
                self._content_block = {"toolUse": deepcopy(start["toolUse"])}
            return

        if "contentBlockDelta" in event:
            # {'contentBlockDelta': {'delta': {'text': "Hello"}, 'contentBlockIndex': 0}}
            # {'contentBlockDelta': {'delta': {'toolUse': {'input': '{"location":"Seattle"}'}}, 'contentBlockIndex': 1}}
            if self._record_message:
                delta = event["contentBlockDelta"].get("delta", {})
                if "text" in delta:
                    self._content_block.setdefault("text", "")
                    self._content_block["text"] += delta["text"]
                elif "toolUse" in delta:
                    self._content_block["toolUse"].setdefault("input", "")
                    self._content_block["toolUse"]["input"] += delta["toolUse"].get(
                        "input", ""
                    )
            return

        if "contentBlockStop" in event:
            # {'contentBlockStop': {'contentBlockIndex': 0}}
            if self._record_message and self._message is not None:
                if "toolUse" in self._content_block:
                    self._content_block["toolUse"] = decode_tool_use_in_stream(
                        self._content_block["toolUse"]
                    )

                self._message["content"].append(self._content_block)
                self._content_block = {}
            return

        if "messageStop" in event:
            # {'messageStop': {'stopReason': 'end_turn'}}
            stop_reason = event["messageStop"].get("stopReason")
            if stop_reason is not None:
                self._response["stopReason"] = stop_reason

            if self._record_message:
                self._response["output"] = {"message": self._message}
                self._record_message = False
                self._message = None

            return

        if "metadata" in event:
            # {'metadata': {'usage': {'inputTokens': 12, 'outputTokens': 15, 'totalTokens': 27}, 'metrics': {'latencyMs': 2980}}}
            usage = event["metadata"].get("usage")
            if usage is not None:
                self._response["usage"] = {}
                input_tokens = usage.get("inputTokens")
                if input_tokens is not None:
                    self._response["usage"]["inputTokens"] = input_tokens

                output_tokens = usage.get("outputTokens")
                if output_tokens is not None:
                    self._response["usage"]["outputTokens"] = output_tokens

            self._stream_done_callback(self._response)

            return



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Optional

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv._incubating.metrics import gen_ai_metrics
from opentelemetry.semconv.attributes import error_attributes as ErrorAttributes

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes
from llm_tracekit.instruments import (
    GEN_AI_CLIENT_OPERATION_DURATION_BUCKETS,
    GEN_AI_CLIENT_TOKEN_USAGE_BUCKETS,
)

# This is a PNG of a single black pixel
IMAGE_DATA = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x01\x00\x00\x00\x007n\xf9$\x00\x00\x00\nIDATx\x01c`\x00\x00\x00\x02\x00\x01su\x01\x18\x00\x00\x00\x00IEND\xaeB`\x82"


def assert_tool_definitions_in_span(span: ReadableSpan, tools: List[dict]):
    assert span.attributes is not None

    for index, tool in enumerate(tools):
        assert (
            span.attributes[
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_NAME.format(
                    tool_index=index
                )
            ]
            == tool["name"]
        )
        assert (
            span.attributes[
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_DESCRIPTION.format(
                    tool_index=index
                )
            ]
            == tool["description"]
        )
        assert (
            span.attributes[
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_PARAMETERS.format(
                    tool_index=index
                )
            ]
            == tool["parameters"]
        )


def assert_attributes_in_span(
    span: ReadableSpan,
    span_name: str,
    request_model: Optional[str] = None,
    response_model: Optional[str] = None,
    usage_input_tokens: Optional[int] = None,
    usage_output_tokens: Optional[int] = None,
    finish_reasons: Optional[Iterable[str]] = None,
    error: Optional[str] = None,
    max_tokens: Optional[int] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    agent_id: Optional[str] = None,
    agent_alias_id: Optional[str] = None,
):
    assert span.name == span_name
    assert span.attributes is not None

    attributes_to_expected_values = {
        GenAIAttributes.GEN_AI_OPERATION_NAME: GenAIAttributes.GenAiOperationNameValues.CHAT.value,
        GenAIAttributes.GEN_AI_SYSTEM: GenAIAttributes.GenAiSystemValues.AWS_BEDROCK.value,
        GenAIAttributes.GEN_AI_REQUEST_MODEL: request_model,
        GenAIAttributes.GEN_AI_RESPONSE_MODEL: response_model,
        GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS: usage_input_tokens,
        GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS: usage_output_tokens,
        GenAIAttributes.GEN_AI_RESPONSE_FINISH_REASONS: finish_reasons,
        ErrorAttributes.ERROR_TYPE: error,
        GenAIAttributes.GEN_AI_REQUEST_MAX_TOKENS: max_tokens,
        GenAIAttributes.GEN_AI_REQUEST_TEMPERATURE: temperature,
        GenAIAttributes.GEN_AI_REQUEST_TOP_P: top_p,
        GenAIAttributes.GEN_AI_AGENT_ID: agent_id,
        ExtendedGenAIAttributes.GEN_AI_BEDROCK_AGENT_ALIAS_ID: agent_alias_id,
    }
    for attribute, expected_value in attributes_to_expected_values.items():
        if expected_value is not None:
            assert span.attributes[attribute] == expected_value, attribute
        else:
            assert attribute not in span.attributes


def assert_expected_metrics(
    metrics,
    usage_input_tokens: Optional[int] = None,
    usage_output_tokens: Optional[int] = None,
    error: Optional[str] = None,
    request_model: Optional[str] = None,
    response_model: Optional[str] = None,
):
    attributes = {
        GenAIAttributes.GEN_AI_OPERATION_NAME: GenAIAttributes.GenAiOperationNameValues.CHAT.value,
        GenAIAttributes.GEN_AI_SYSTEM: GenAIAttributes.GenAiSystemValues.AWS_BEDROCK.value,
        GenAIAttributes.GEN_AI_REQUEST_MODEL: request_model,
        GenAIAttributes.GEN_AI_RESPONSE_MODEL: response_model,
        ErrorAttributes.ERROR_TYPE: error,
    }

    metric_data_points = []
    duration_metric = None
    usage_metric = None
    for metric in metrics:
        if metric.name == gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION:
            duration_metric = metric
        elif metric.name == gen_ai_metrics.GEN_AI_CLIENT_TOKEN_USAGE:
            usage_metric = metric

    assert duration_metric is not None
    assert duration_metric.data.data_points[0].sum > 0
    assert (
        list(duration_metric.data.data_points[0].explicit_bounds)
        == GEN_AI_CLIENT_OPERATION_DURATION_BUCKETS
    )
    metric_data_points.append(duration_metric.data.data_points[0])

    if usage_input_tokens is not None:
        assert usage_metric is not None
        input_token_usage = next(
            (
                data_point
                for data_point in usage_metric.data.data_points
                if data_point.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
                == GenAIAttributes.GenAiTokenTypeValues.INPUT.value
            ),
            None,
        )
        assert input_token_usage is not None
        assert input_token_usage.sum == usage_input_tokens
        assert (
            list(input_token_usage.explicit_bounds) == GEN_AI_CLIENT_TOKEN_USAGE_BUCKETS
        )
        metric_data_points.append(input_token_usage)

    if usage_output_tokens is not None:
        assert usage_metric is not None
        output_token_usage = next(
            (
                data_point
                for data_point in usage_metric.data.data_points
                if data_point.attributes[GenAIAttributes.GEN_AI_TOKEN_TYPE]
                == GenAIAttributes.GenAiTokenTypeValues.OUTPUT.value
            ),
            None,
        )
        assert output_token_usage is not None
        assert output_token_usage.sum == usage_output_tokens
        assert (
            list(output_token_usage.explicit_bounds)
            == GEN_AI_CLIENT_TOKEN_USAGE_BUCKETS
        )
        metric_data_points.append(output_token_usage)

    # Assert that all data points have all the expected attributes
    for data_point in metric_data_points:
        for attribute, expected_value in attributes.items():
            if expected_value is not None:
                assert data_point.attributes[attribute] == expected_value
            else:
                assert attribute not in data_point.attributes



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import defaultdict
from contextlib import suppress
from enum import Enum
from timeit import default_timer
from typing import Any, Callable, Dict, List, Optional, Union

from botocore.eventstream import EventStream, EventStreamError
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.trace import Span
from wrapt import ObjectProxy

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes
from llm_tracekit.bedrock.utils import decode_tool_use_in_stream, record_metrics
from llm_tracekit.instruments import Instruments
from llm_tracekit.span_builder import (
    Choice,
    Message,
    ToolCall,
    attribute_generator,
    generate_base_attributes,
    generate_choice_attributes,
    generate_message_attributes,
    generate_request_attributes,
    generate_response_attributes,
)


class _ModelType(Enum):
    CLAUDE = "claude"
    LLAMA3 = "llama3"


def _get_model_type_from_model_id(model_id: Optional[str]) -> Optional[_ModelType]:
    if model_id is None:
        return None

    if "anthropic.claude" in model_id:
        return _ModelType.CLAUDE
    elif "meta.llama3" in model_id:
        return _ModelType.LLAMA3

    return None


def _parse_claude_message(
    role: Optional[str], content: Optional[Union[str, list]]
) -> List[Message]:
    """Attempts to combine the content parts of a `converse` API message to a single message."""
    if isinstance(content, str):
        return [Message(role=role, content=content)]

    if isinstance(content, list):
        content_blocks_by_type = defaultdict(list)
        for block in content:
            content_blocks_by_type[block.get("type")].append(block)


        # We follow the same logic the OTEL implementation uses:
        #  * For assistant messages (text/tool calls) - return a single message
        #  * for user messages (text / tool call results) - return a message for each tool 
        #    call result, and another message for text
        messages = []
        if role == "assistant":
            content = None
            tool_calls = None
            if "tool_use" in content_blocks_by_type:
                tool_calls = []
                for block in content_blocks_by_type["tool_use"]:
                    arguments = None
                    if "input" in block:
                        arguments = json.dumps(block["input"])

                    tool_calls.append(
                        ToolCall(
                            id=block.get("id"),
                            type="function",
                            function_name=block.get("name"),
                            function_arguments=arguments,
                        )
                    )
            
            if "text" in content_blocks_by_type:
                text_parts = [
                    block["text"]
                    for block in content_blocks_by_type["text"]
                    if block.get("text") is not None
                ]
                content = "".join(text_parts)

            messages.append(
                Message(
                    role=role,
                    tool_calls=tool_calls,
                    content=content,
                )
            )
        elif role == "user":
            if "tool_result" in content_blocks_by_type:
                for tool_result in content_blocks_by_type["tool_result"]:
                    tool_result_content = tool_result.get("content")
                    if not isinstance(tool_result_content, str):
                        tool_result_content = None

                    messages.append(
                        Message(
                            role="tool",
                            tool_call_id=tool_result.get("tool_use_id"),
                            content=tool_result_content,
                        )
                    )
            if "text" in content_blocks_by_type:
                text_parts = [
                    block["text"]
                    for block in content_blocks_by_type["text"]
                    if block.get("text") is not None
                ]
                messages.append(Message(role=role, content="".join(text_parts)))

        if len(messages) > 0:
            return messages

    return [Message(role=role)]


@attribute_generator
def _generate_claude_request_and_message_attributes(
    model_id: Optional[str], parsed_body: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    messages = []
    if "system" in parsed_body and isinstance(parsed_body["system"], str):
        messages.append(Message(role="system", content=parsed_body["system"]))

    for message in parsed_body.get("messages", []):
        messages.extend(
            _parse_claude_message(
                role=message.get("role"), content=message.get("content")
            )
        )

    tool_attributes = {}
    tools = parsed_body.get("tools", [])
    for index, tool_definition in enumerate(tools):
        tool_params = None
        if (
            "input_schema" in tool_definition
            and "properties" in tool_definition["input_schema"]
        ):
            with suppress(TypeError):
                tool_params = json.dumps(tool_definition["input_schema"])

        tool_attributes.update(
            {
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_NAME.format(
                    tool_index=index
                ): tool_definition.get("name"),
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_DESCRIPTION.format(
                    tool_index=index
                ): tool_definition.get("description"),
                ExtendedGenAIAttributes.GEN_AI_BEDROCK_REQUEST_TOOLS_FUNCTION_PARAMETERS.format(
                    tool_index=index
                ): tool_params,
            }
        )

    return {
        **generate_request_attributes(
            model=model_id,
            max_tokens=parsed_body.get("max_tokens"),
            temperature=parsed_body.get("temperature"),
            top_p=parsed_body.get("top_p"),
        ),
        **generate_message_attributes(
            messages=messages, capture_content=capture_content
        ),
        **tool_attributes,
    }


def _generate_claude_response_and_choice_attributes(
    parsed_body: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    finish_reason = parsed_body.get("stop_reason")
    usage_data = parsed_body.get("usage", {})
    parsed_response_message = _parse_claude_message(
        role=parsed_body.get("role"), content=parsed_body.get("content")
    )[0]
    choice = Choice(
        finish_reason=finish_reason,
        role=parsed_response_message.role,
        content=parsed_response_message.content,
        tool_calls=parsed_response_message.tool_calls,
    )

    return {
        **generate_response_attributes(
            model=parsed_body.get("model"),
            finish_reasons=[] if finish_reason is None else [finish_reason],
            id=parsed_body.get("id"),
            usage_input_tokens=usage_data.get("input_tokens"),
            usage_output_tokens=usage_data.get("output_tokens"),
        ),
        **generate_choice_attributes(choices=[choice], capture_content=capture_content),
    }


def _generate_llama_request_and_message_attributes(
    model_id: Optional[str], parsed_body: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    attributes = generate_request_attributes(
        model=model_id,
        max_tokens=parsed_body.get("max_gen_len"),
        temperature=parsed_body.get("temperature"),
        top_p=parsed_body.get("top_p"),
    )
    if "prompt" in parsed_body:
        messages = [Message(role="user", content=parsed_body["prompt"])]
        attributes.update(
            generate_message_attributes(
                messages=messages, capture_content=capture_content
            )
        )

    return attributes


def _generate_llama_response_and_choice_attributes(
    model_id: Optional[str], parsed_body: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    finish_reason = parsed_body.get("stop_reason")
    usage_input_tokens = parsed_body.get("prompt_token_count")
    usage_output_tokens = parsed_body.get("generation_token_count")
    return {
        **generate_response_attributes(
            model=model_id,
            finish_reasons=None if finish_reason is None else [finish_reason],
            usage_input_tokens=usage_input_tokens,
            usage_output_tokens=usage_output_tokens,
        ),
        **generate_choice_attributes(
            choices=[
                Choice(
                    finish_reason=finish_reason,
                    role="assistant",
                    content=parsed_body.get("generation"),
                )
            ],
            capture_content=capture_content,
        ),
    }


def generate_attributes_from_invoke_input(
    kwargs: Dict[str, Any], capture_content: bool
) -> Dict[str, Any]:
    base_attributes = generate_base_attributes(
        system=GenAIAttributes.GenAiSystemValues.AWS_BEDROCK
    )
    model_id = kwargs.get("modelId")
    model_type = _get_model_type_from_model_id(model_id)
    partial_attributes = {
        **base_attributes,
        **generate_request_attributes(model=model_id),
    }

    if model_type is None:
        return partial_attributes

    body = kwargs.get("body")
    if body is None:
        return partial_attributes

    try:
        parsed_body = json.loads(body)
    except json.JSONDecodeError:
        return partial_attributes

    if model_type is _ModelType.CLAUDE:
        return {
            **base_attributes,
            **_generate_claude_request_and_message_attributes(
                model_id=model_id,
                parsed_body=parsed_body,
                capture_content=capture_content,
            ),
        }
    elif model_type is _ModelType.LLAMA3:
        return {
            **base_attributes,
            **_generate_llama_request_and_message_attributes(
                model_id=model_id,
                parsed_body=parsed_body,
                capture_content=capture_content,
            ),
        }

    return partial_attributes


def record_invoke_model_result_attributes(
    result_body: Union[str, Dict[str, Any]],
    span: Span,
    start_time: float,
    instruments: Instruments,
    capture_content: bool,
    model_id: Optional[str],
):
    request_model = model_id
    response_model = model_id
    usage_input_tokens = None
    usage_output_tokens = None
    try:
        model_type = _get_model_type_from_model_id(model_id)
        if model_type is None:
            return

        parsed_body = {}
        if isinstance(result_body, dict):
            parsed_body = result_body
        elif isinstance(result_body, (str, bytes)):
            try:
                parsed_body = json.loads(result_body)
            except json.JSONDecodeError:
                return

        if model_type is _ModelType.LLAMA3:
            span.set_attributes(
                _generate_llama_response_and_choice_attributes(
                    model_id=model_id,
                    parsed_body=parsed_body,
                    capture_content=capture_content,
                )
            )
            usage_input_tokens = parsed_body.get("prompt_token_count")
            usage_output_tokens = parsed_body.get("generation_token_count")
        elif model_type is _ModelType.CLAUDE:
            span.set_attributes(
                _generate_claude_response_and_choice_attributes(
                    parsed_body=parsed_body, capture_content=capture_content
                )
            )
            response_model = parsed_body.get("model")
            usage_input_tokens = parsed_body.get("usage", {}).get("input_tokens")
            usage_output_tokens = parsed_body.get("usage", {}).get("output_tokens")

    finally:
        duration = max((default_timer() - start_time), 0)
        span.end()
        record_metrics(
            instruments=instruments,
            duration=duration,
            request_model=request_model,
            response_model=response_model,
            usage_input_tokens=usage_input_tokens,
            usage_output_tokens=usage_output_tokens,
        )


class InvokeModelWithResponseStreamWrapper(ObjectProxy):
    """Wrapper for botocore.eventstream.EventStream"""

    def __init__(
        self,
        stream: EventStream,
        stream_done_callback: Callable[[Dict[str, Union[int, str]]], None],
        stream_error_callback: Callable[[Exception], None],
        model_id: Optional[str],
    ):
        super().__init__(stream)

        self._stream_done_callback = stream_done_callback
        self._stream_error_callback = stream_error_callback
        self._model_id = model_id

        # accumulating things in the same shape of the Converse API
        # {"usage": {"inputTokens": 0, "outputTokens": 0}, "stopReason": "finish", "output": {"message": {"role": "", "content": [{"text": ""}]}
        self._response: Dict[str, Any] = {}
        self._message = None
        self._content_block: Dict[str, Any] = {}
        self._tool_json_input_buffer = ""
        self._record_message = False

    def __iter__(self):
        try:
            for event in self.__wrapped__:
                self._process_event(event)
                yield event
        except EventStreamError as exc:
            self._stream_error_callback(exc)
            raise

    def _process_event(self, event):
        if "chunk" not in event:
            return

        json_bytes = event["chunk"].get("bytes", b"")
        decoded = json_bytes.decode("utf-8")
        try:
            chunk = json.loads(decoded)
        except json.JSONDecodeError:
            return

        model_type = _get_model_type_from_model_id(self._model_id)

        if model_type is _ModelType.LLAMA3:
            self._process_meta_llama_chunk(chunk)
        elif model_type is _ModelType.CLAUDE:
            self._process_anthropic_claude_chunk(chunk)

    def _process_meta_llama_invocation_metrics(self, invocation_metrics):
        input_tokens = invocation_metrics.get("inputTokenCount")
        if input_tokens is not None:
            self._response["prompt_token_count"] = input_tokens

        output_tokens = invocation_metrics.get("outputTokenCount")
        if output_tokens is not None:
            self._response["generation_token_count"] = output_tokens

    def _process_anthropic_claude_invocation_metrics(self, invocation_metrics):
        self._response["usage"] = {}
        input_tokens = invocation_metrics.get("inputTokenCount")
        if input_tokens is not None:
            self._response["usage"]["input_tokens"] = input_tokens

        output_tokens = invocation_metrics.get("outputTokenCount")
        if output_tokens is not None:
            self._response["usage"]["output_tokens"] = output_tokens

    def _process_meta_llama_chunk(self, chunk):
        if self._message is None:
            self._message = {"generation": ""}

        self._message["generation"] += chunk.get("generation", "")

        if chunk.get("stop_reason") is not None and self._message is not None:
            invocation_metrics = chunk.get("amazon-bedrock-invocationMetrics")
            if invocation_metrics is not None:
                self._process_meta_llama_invocation_metrics(invocation_metrics)

            self._response["stop_reason"] = chunk["stop_reason"]
            self._response["generation"] = self._message["generation"]
            self._message = None

            self._stream_done_callback(self._response)
            return

    def _process_anthropic_claude_chunk(self, chunk):
        # pylint: disable=too-many-return-statements,too-many-branches
        message_type = chunk.get("type")
        if message_type is None:
            return

        if message_type == "message_start":
            # {'type': 'message_start', 'message': {'id': 'id', 'type': 'message', 'role': 'assistant', 'model': 'claude-2.0', 'content': [], 'stop_reason': None, 'stop_sequence': None, 'usage': {'input_tokens': 18, 'output_tokens': 1}}}
            if chunk.get("message", {}).get("role") == "assistant":
                self._record_message = True
                message = chunk["message"]
                self._message = {
                    "model": message["model"],
                    "role": message["role"],
                    "content": message.get("content", []),
                }
            return

        if message_type == "content_block_start":
            # {'type': 'content_block_start', 'index': 0, 'content_block': {'type': 'text', 'text': ''}}
            # {'type': 'content_block_start', 'index': 1, 'content_block': {'type': 'tool_use', 'id': 'id', 'name': 'func_name', 'input': {}}}
            if self._record_message:
                block = chunk.get("content_block", {})
                if block.get("type") == "text":
                    self._content_block = block
                elif block.get("type") == "tool_use":
                    self._content_block = block
            return

        if message_type == "content_block_delta":
            # {'type': 'content_block_delta', 'index': 0, 'delta': {'type': 'text_delta', 'text': 'Here'}}
            # {'type': 'content_block_delta', 'index': 1, 'delta': {'type': 'input_json_delta', 'partial_json': ''}}
            if self._record_message:
                delta = chunk.get("delta", {})
                if delta.get("type") == "text_delta":
                    self._content_block["text"] += delta.get("text", "")
                elif delta.get("type") == "input_json_delta":
                    self._tool_json_input_buffer += delta.get("partial_json", "")
            return

        if message_type == "content_block_stop":
            # {'type': 'content_block_stop', 'index': 0}
            if self._tool_json_input_buffer:
                self._content_block["input"] = self._tool_json_input_buffer

            if self._message is not None:
                self._message["content"].append(
                    decode_tool_use_in_stream(self._content_block)
                )
            self._content_block = {}
            self._tool_json_input_buffer = ""
            return

        if message_type == "message_delta":
            # {'type': 'message_delta', 'delta': {'stop_reason': 'end_turn', 'stop_sequence': None}, 'usage': {'output_tokens': 123}}
            stop_reason = chunk.get("delta", {}).get("stop_reason")
            if stop_reason is not None:
                self._response["stop_reason"] = stop_reason
            return

        if message_type == "message_stop":
            # {'type': 'message_stop', 'amazon-bedrock-invocationMetrics': {'inputTokenCount': 18, 'outputTokenCount': 123, 'invocationLatency': 5250, 'firstByteLatency': 290}}
            invocation_metrics = chunk.get("amazon-bedrock-invocationMetrics")
            if invocation_metrics is not None:
                self._process_anthropic_claude_invocation_metrics(invocation_metrics)

            if self._record_message and self._message is not None:
                self._response["model"] = self._message["model"]
                self._response["role"] = self._message["role"]
                self._response["content"] = self._message["content"]
                self._record_message = False
                self._message = None

            self._stream_done_callback(self._response)
            return



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opentelemetry.metrics import Histogram, Meter
from opentelemetry.semconv._incubating.metrics import gen_ai_metrics

GEN_AI_CLIENT_OPERATION_DURATION_BUCKETS = [
    0.01,
    0.02,
    0.04,
    0.08,
    0.16,
    0.32,
    0.64,
    1.28,
    2.56,
    5.12,
    10.24,
    20.48,
    40.96,
    81.92,
]

GEN_AI_CLIENT_TOKEN_USAGE_BUCKETS = [
    1,
    4,
    16,
    64,
    256,
    1024,
    4096,
    16384,
    65536,
    262144,
    1048576,
    4194304,
    16777216,
    67108864,
]


class Instruments:
    def __init__(self, meter: Meter):
        self.operation_duration_histogram: Histogram = meter.create_histogram(
            name=gen_ai_metrics.GEN_AI_CLIENT_OPERATION_DURATION,
            description="GenAI operation duration",
            unit="s",
            explicit_bucket_boundaries_advisory=GEN_AI_CLIENT_OPERATION_DURATION_BUCKETS,
        )
        self.token_usage_histogram: Histogram = meter.create_histogram(
            name=gen_ai_metrics.GEN_AI_CLIENT_TOKEN_USAGE,
            description="Measures number of input and output tokens used",
            unit="{token}",
            explicit_bucket_boundaries_advisory=GEN_AI_CLIENT_TOKEN_USAGE_BUCKETS,
        )



# Copyright Coralogix Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json

import boto3
import pytest
from botocore.exceptions import ClientError

from tests.bedrock.utils import (
    IMAGE_DATA,
    assert_attributes_in_span,
    assert_expected_metrics,
    assert_tool_definitions_in_span,
)
from tests.utils import assert_choices_in_span, assert_messages_in_span


def _get_current_weather_tool_definition():
    return {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "input_schema": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. Boston, MA",
                },
            },
            "required": ["location"],
            "additionalProperties": False,
        },
    }


def _convert_claude_stream_to_response(stream) -> dict:
    result = {
        "model": "",
        "role": "",
        "content": [],
        "stop_reason": "",
        "usage": {
            "input_tokens": 0,
            "output_tokens": 0,
        },
    }

    current_block = {}
    for event in stream:
        parsed_chunk = json.loads(event["chunk"]["bytes"])
        if parsed_chunk["type"] == "message_start":
            result["model"] = parsed_chunk["message"]["model"]
            result["role"] = parsed_chunk["message"]["role"]
        elif parsed_chunk["type"] == "content_block_start":
            current_block = parsed_chunk["content_block"]
            if "input" in current_block:
                current_block["input"] = ""
        elif parsed_chunk["type"] == "content_block_delta":
            if parsed_chunk["delta"]["type"] == "text_delta":
                current_block["text"] += parsed_chunk["delta"]["text"]
            elif parsed_chunk["delta"]["type"] == "input_json_delta":
                current_block["input"] += parsed_chunk["delta"]["partial_json"]
        if parsed_chunk["type"] == "content_block_stop":
            if current_block["type"] == "tool_use":
                current_block["input"] = json.loads(current_block["input"])

            result["content"].append(current_block)
            current_block = {}
        elif parsed_chunk["type"] == "message_delta":
            result["stop_reason"] = parsed_chunk["delta"]["stop_reason"]
        elif parsed_chunk["type"] == "message_stop":
            result["usage"]["input_tokens"] = parsed_chunk[
                "amazon-bedrock-invocationMetrics"
            ]["inputTokenCount"]
            result["usage"]["output_tokens"] = parsed_chunk[
                "amazon-bedrock-invocationMetrics"
            ]["outputTokenCount"]

    return result


def _run_and_check_invoke_model_llama(
    bedrock_client,
    model_id: str,
    span_exporter,
    metric_reader,
    expect_content: bool,
    stream: bool,
):
    args = {
        "modelId": model_id,
        "contentType": "application/json",
        "accept": "application/json",
        "body": json.dumps(
            {
                "prompt": "say this is a test",
                "temperature": 0,
                "max_gen_len": 300,
                "top_p": 1,
            }
        ),
    }
    if stream:
        span_name = "bedrock.invoke_model_with_response_stream"
        stream_result = bedrock_client.invoke_model_with_response_stream(**args)
        result = {
            "stop_reason": "",
            "generation": "",
            "prompt_token_count": 0,
            "generation_token_count": 0,
        }
        for chunk in stream_result["body"]:
            parsed_chunk = json.loads(chunk["chunk"]["bytes"])
            result["generation"] += parsed_chunk.get("generation", "")
            if parsed_chunk.get("stop_reason") is not None:
                result["stop_reason"] = parsed_chunk["stop_reason"]
                result["prompt_token_count"] = parsed_chunk[
                    "amazon-bedrock-invocationMetrics"
                ]["inputTokenCount"]
                result["generation_token_count"] = parsed_chunk[
                    "amazon-bedrock-invocationMetrics"
                ]["outputTokenCount"]
    else:
        span_name = "bedrock.invoke_model"
        invoke_result = bedrock_client.invoke_model(**args)
        result = json.loads(invoke_result["body"].read())

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name=span_name,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result["prompt_token_count"],
        usage_output_tokens=result["generation_token_count"],
        finish_reasons=(result["stop_reason"],),
        max_tokens=300,
        temperature=0,
        top_p=1,
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages,
        expect_content=expect_content,
    )

    expected_choice = {
        "finish_reason": result["stop_reason"],
        "message": {
            "role": "assistant",
            "content": result["generation"],
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[expected_choice], expect_content=expect_content
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        response_model=model_id,
        usage_input_tokens=result["prompt_token_count"],
        usage_output_tokens=result["generation_token_count"],
    )


def _run_and_check_invoke_model_claude(
    bedrock_client,
    model_id: str,
    span_exporter,
    metric_reader,
    expect_content: bool,
    stream: bool,
):
    args = {
        "modelId": model_id,
        "contentType": "application/json",
        "accept": "application/json",
        "body": json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 300,
                "messages": [{"role": "user", "content": "say this is a test"}],
                "system": "You are a helpful assistant",
                "temperature": 0,
                "top_p": 1,
            }
        ),
    }
    if stream:
        span_name = "bedrock.invoke_model_with_response_stream"
        stream_result = bedrock_client.invoke_model_with_response_stream(**args)
        result = _convert_claude_stream_to_response(stream_result["body"])
    else:
        span_name = "bedrock.invoke_model"
        invoke_result = bedrock_client.invoke_model(**args)
        result = json.loads(invoke_result["body"].read())

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name=span_name,
        request_model=model_id,
        response_model=result["model"],
        usage_input_tokens=result["usage"]["input_tokens"],
        usage_output_tokens=result["usage"]["output_tokens"],
        finish_reasons=(result["stop_reason"],),
        max_tokens=300,
        temperature=0,
        top_p=1,
    )

    expected_messages = [
        {"role": "system", "content": "You are a helpful assistant"},
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages,
        expect_content=expect_content,
    )

    expected_choice = {
        "finish_reason": result["stop_reason"],
        "message": {
            "role": result["role"],
            "content": result["content"][0]["text"],
        },
    }
    assert_choices_in_span(
        span=spans[0], expected_choices=[expected_choice], expect_content=expect_content
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        response_model=result["model"],
        usage_input_tokens=result["usage"]["input_tokens"],
        usage_output_tokens=result["usage"]["output_tokens"],
    )


def _run_and_check_invoke_model_claude_tool_calls(
    bedrock_client,
    model_id: str,
    span_exporter,
    metric_reader,
    expect_content: bool,
    stream: bool,
):
    tool_definition = _get_current_weather_tool_definition()
    common_args = {
        "modelId": model_id,
        "contentType": "application/json",
        "accept": "application/json",
    }
    body = {
        "anthropic_version": "bedrock-2023-05-31",
        "max_tokens": 1000,
        "messages": [
            {
                "role": "user",
                "content": "What's the weather in Seattle and San Francisco today?",
            }
        ],
        "system": "You are a helpful assistant",
        "tools": [tool_definition],
        "tool_choice": {"type": "auto"},
    }
    args = {
        **common_args,
        "body": json.dumps(body),
    }
    if stream:
        span_name = "bedrock.invoke_model_with_response_stream"
        stream_result_0 = bedrock_client.invoke_model_with_response_stream(**args)
        result_0 = _convert_claude_stream_to_response(stream_result_0["body"])
    else:
        span_name = "bedrock.invoke_model"
        invoke_result_0 = bedrock_client.invoke_model(**args)
        result_0 = json.loads(invoke_result_0["body"].read())

    tool_call_blocks = [
        block for block in result_0["content"] if block["type"] == "tool_use"
    ]

    assert result_0["stop_reason"] == "tool_use"

    tool_results = [
        {
            "tool_call_id": tool_call_blocks[0]["id"],
            "content": "50 degrees and raining",
        },
        {"tool_call_id": tool_call_blocks[1]["id"], "content": "70 degrees and sunny"},
    ]
    expected_assistant_message = {
        "role": result_0["role"],
        "content": result_0["content"][0]["text"],
        "tool_calls": [
            {
                "id": tool_call_blocks[0]["id"],
                "type": "function",
                "function": {
                    "name": tool_call_blocks[0]["name"],
                    "arguments": json.dumps(tool_call_blocks[0]["input"]),
                },
            },
            {
                "id": tool_call_blocks[1]["id"],
                "type": "function",
                "function": {
                    "name": tool_call_blocks[1]["name"],
                    "arguments": json.dumps(tool_call_blocks[1]["input"]),
                },
            },
        ],
    }

    body["messages"].append({"role": result_0["role"], "content": result_0["content"]})
    body["messages"].append(
        {
            "role": "user",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": tool_results[0]["tool_call_id"],
                    "content": tool_results[0]["content"],
                },
                {
                    "type": "tool_result",
                    "tool_use_id": tool_results[1]["tool_call_id"],
                    "content": tool_results[1]["content"],
                },
            ],
        }
    )
    args = {
        **common_args,
        "body": json.dumps(body),
    }

    if stream:
        stream_result_1 = bedrock_client.invoke_model_with_response_stream(**args)
        result_1 = _convert_claude_stream_to_response(stream_result_1["body"])
    else:
        invoke_result_1 = bedrock_client.invoke_model(**args)
        result_1 = json.loads(invoke_result_1["body"].read())

    assert result_1["stop_reason"] == "end_turn"

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 2

    assert_attributes_in_span(
        span=spans[0],
        span_name=span_name,
        request_model=model_id,
        response_model=result_0["model"],
        usage_input_tokens=result_0["usage"]["input_tokens"],
        usage_output_tokens=result_0["usage"]["output_tokens"],
        finish_reasons=(result_0["stop_reason"],),
        max_tokens=1000,
    )
    assert_attributes_in_span(
        span=spans[1],
        span_name=span_name,
        request_model=model_id,
        response_model=result_1["model"],
        usage_input_tokens=result_1["usage"]["input_tokens"],
        usage_output_tokens=result_1["usage"]["output_tokens"],
        finish_reasons=(result_1["stop_reason"],),
        max_tokens=1000,
    )

    expected_tool_definition = {
        "name": tool_definition["name"],
        "description": tool_definition["description"],
        "parameters": json.dumps(tool_definition["input_schema"]),
    }
    assert_tool_definitions_in_span(spans[0], [expected_tool_definition])
    assert_tool_definitions_in_span(spans[1], [expected_tool_definition])

    expected_messages_0 = [
        {"role": "system", "content": "You are a helpful assistant"},
        {
            "role": "user",
            "content": "What's the weather in Seattle and San Francisco today?",
        },
    ]
    assert_messages_in_span(
        span=spans[0],
        expected_messages=expected_messages_0,
        expect_content=expect_content,
    )

    expected_messages_1 = [
        *expected_messages_0,
        expected_assistant_message,
        {"role": "tool", **tool_results[0]},
        {"role": "tool", **tool_results[1]},
    ]
    assert_messages_in_span(
        span=spans[1],
        expected_messages=expected_messages_1,
        expect_content=expect_content,
    )

    expected_choice_0 = {
        "finish_reason": result_0["stop_reason"],
        "message": expected_assistant_message,
    }
    assert_choices_in_span(
        span=spans[0],
        expected_choices=[expected_choice_0],
        expect_content=expect_content,
    )
    expected_choice_1 = {
        "finish_reason": result_1["stop_reason"],
        "message": {
            "role": result_1["role"],
            "content": result_1["content"][0]["text"],
        },
    }
    assert_choices_in_span(
        span=spans[1],
        expected_choices=[expected_choice_1],
        expect_content=expect_content,
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        response_model=result_0["model"],
        usage_input_tokens=result_0["usage"]["input_tokens"]
        + result_1["usage"]["input_tokens"],
        usage_output_tokens=result_0["usage"]["output_tokens"]
        + result_1["usage"]["output_tokens"],
    )


@pytest.mark.vcr()
def test_invoke_model_calude_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_calude_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_calude_tool_calls_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude_tool_calls(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_calude_tool_calls_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude_tool_calls(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_calude_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    bedrock_client_with_content.invoke_model(
        modelId=claude_model_id,
        contentType="application/json",
        accept="application/json",
        body=json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 300,
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "say this "},
                            {"type": "text", "text": "is a test"},
                        ],
                    }
                ],
            }
        ),
    )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_invoke_model_calude_unsupported_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    bedrock_client_with_content.invoke_model(
        modelId=claude_model_id,
        contentType="application/json",
        accept="application/json",
        body=json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 300,
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "say this "},
                            {
                                "type": "image",
                                "source": {
                                    "type": "base64",
                                    "media_type": "image/png",
                                    "data": base64.b64encode(IMAGE_DATA).decode(),
                                },
                            },
                            {"type": "text", "text": "is a test"},
                        ],
                    }
                ],
            }
        ),
    )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_invoke_model_llama_with_content(
    bedrock_client_with_content, llama_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_llama(
        bedrock_client=bedrock_client_with_content,
        model_id=llama_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_llama_no_content(
    bedrock_client_no_content, llama_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_llama(
        bedrock_client=bedrock_client_no_content,
        model_id=llama_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=False,
    )


@pytest.mark.vcr()
def test_invoke_model_non_existing_model(
    bedrock_client_with_content, span_exporter, metric_reader
):
    model_id = "meta.llama3-8b-instruct-v999:999"
    with pytest.raises(Exception):
        bedrock_client_with_content.invoke_model(
            modelId=model_id,
            contentType="application/json",
            accept="application/json",
            body=json.dumps({"prompt": "say this is a test"}),
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_model",
        request_model=model_id,
        error="ValidationException",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        error="ValidationException",
    )


@pytest.mark.vcr()
def test_invoke_model_bad_auth(
    instrument_with_content, llama_model_id: str, span_exporter, metric_reader
):
    client = boto3.client(
        "bedrock-runtime",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        aws_session_token="test",
    )
    with pytest.raises(ClientError):
        client.invoke_model(
            modelId=llama_model_id,
            contentType="application/json",
            accept="application/json",
            body=json.dumps({"prompt": "say this is a test"}),
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_model",
        request_model=llama_model_id,
        error="ClientError",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=llama_model_id,
        error="ClientError",
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_tool_calls_with_content(
    bedrock_client_with_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude_tool_calls(
        bedrock_client=bedrock_client_with_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_tool_calls_no_content(
    bedrock_client_no_content, claude_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_claude_tool_calls(
        bedrock_client=bedrock_client_no_content,
        model_id=claude_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    result = bedrock_client_with_content.invoke_model_with_response_stream(
        modelId=claude_model_id,
        contentType="application/json",
        accept="application/json",
        body=json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 300,
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "say this "},
                            {"type": "text", "text": "is a test"},
                        ],
                    }
                ],
            }
        ),
    )

    # consume the stream
    for _ in result["body"]:
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_calude_unsupported_content_blocks(
    bedrock_client_with_content, claude_model_id: str, span_exporter
):
    result = bedrock_client_with_content.invoke_model_with_response_stream(
        modelId=claude_model_id,
        contentType="application/json",
        accept="application/json",
        body=json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 300,
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "say this "},
                            {
                                "type": "image",
                                "source": {
                                    "type": "base64",
                                    "media_type": "image/png",
                                    "data": base64.b64encode(IMAGE_DATA).decode(),
                                },
                            },
                            {"type": "text", "text": "is a test"},
                        ],
                    }
                ],
            }
        ),
    )

    # consume the stream
    for _ in result["body"]:
        pass

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_llama_with_content(
    bedrock_client_with_content, llama_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_llama(
        bedrock_client=bedrock_client_with_content,
        model_id=llama_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=True,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_llama_no_content(
    bedrock_client_no_content, llama_model_id: str, span_exporter, metric_reader
):
    _run_and_check_invoke_model_llama(
        bedrock_client=bedrock_client_no_content,
        model_id=llama_model_id,
        span_exporter=span_exporter,
        metric_reader=metric_reader,
        expect_content=False,
        stream=True,
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_non_existing_model(
    bedrock_client_with_content, span_exporter, metric_reader
):
    model_id = "meta.llama3-8b-instruct-v999:999"
    with pytest.raises(Exception):
        bedrock_client_with_content.invoke_model_with_response_stream(
            modelId=model_id,
            contentType="application/json",
            accept="application/json",
            body=json.dumps({"prompt": "say this is a test"}),
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_model_with_response_stream",
        request_model=model_id,
        error="ValidationException",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=model_id,
        error="ValidationException",
    )


@pytest.mark.vcr()
def test_invoke_model_with_response_stream_bad_auth(
    instrument_with_content, llama_model_id: str, span_exporter, metric_reader
):
    client = boto3.client(
        "bedrock-runtime",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        aws_session_token="test",
    )
    with pytest.raises(ClientError):
        client.invoke_model_with_response_stream(
            modelId=llama_model_id,
            contentType="application/json",
            accept="application/json",
            body=json.dumps({"prompt": "say this is a test"}),
        )

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1

    assert_attributes_in_span(
        span=spans[0],
        span_name="bedrock.invoke_model_with_response_stream",
        request_model=llama_model_id,
        error="ClientError",
    )

    expected_messages = [
        {"role": "user", "content": "say this is a test"},
    ]
    assert_messages_in_span(
        span=spans[0], expected_messages=expected_messages, expect_content=True
    )

    metrics = metric_reader.get_metrics_data().resource_metrics
    assert len(metrics) == 1

    metric_data = metrics[0].scope_metrics[0].metrics
    assert_expected_metrics(
        metrics=metric_data,
        request_model=llama_model_id,
        error="ClientError",
    )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
OpenAI client instrumentation supporting `openai`, it can be enabled by
using ``OpenAIInstrumentor``.

.. _openai: https://pypi.org/project/openai/

Usage
-----

.. code:: python

    from openai import OpenAI
    from llm_tracekit import OpenAIInstrumentor

    OpenAIInstrumentor().instrument()

    client = OpenAI()
    response = client.chat.completions.create(
        model="gpt-4o-mini",
        messages=[
            {"role": "user", "content": "Write a short poem on open telemetry."},
        ],
    )

API
---
"""

from typing import Collection

import openai
from opentelemetry.instrumentation.instrumentor import (  # type: ignore[attr-defined] # Mypy doesn't recognize the attribute
    BaseInstrumentor,
)
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.metrics import get_meter
from opentelemetry.semconv.schemas import Schemas
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from llm_tracekit.instrumentation_utils import is_content_enabled
from llm_tracekit.instruments import Instruments
from llm_tracekit.openai.package import _instruments
from llm_tracekit.openai.patch import (
    async_chat_completions_create,
    chat_completions_create,
)


class OpenAIInstrumentor(BaseInstrumentor):
    def __init__(self):
        self._meter = None

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        """Enable OpenAI instrumentation."""
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(
            __name__,
            "",
            tracer_provider,
            schema_url=Schemas.V1_28_0.value,
        )
        meter_provider = kwargs.get("meter_provider")
        self._meter = get_meter(
            __name__,
            "",
            meter_provider,
            schema_url=Schemas.V1_28_0.value,
        )

        instruments = Instruments(self._meter)

        wrap_function_wrapper(
            module="openai.resources.chat.completions",
            name="Completions.create",
            wrapper=chat_completions_create(tracer, instruments, is_content_enabled()),
        )

        wrap_function_wrapper(
            module="openai.resources.chat.completions",
            name="AsyncCompletions.create",
            wrapper=async_chat_completions_create(
                tracer, instruments, is_content_enabled()
            ),
        )

    def _uninstrument(self, **kwargs):
        unwrap(openai.resources.chat.completions.Completions, "create")
        unwrap(openai.resources.chat.completions.AsyncCompletions, "create")



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from openai.types.chat import ChatCompletion
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv._incubating.attributes import (
    server_attributes as ServerAttributes,
)


def assert_completion_attributes(
    span: ReadableSpan,
    request_model: str,
    response: ChatCompletion,
    operation_name: str = "chat",
    server_address: str = "api.openai.com",
):
    return assert_all_attributes(
        span,
        request_model,
        response.id,
        response.model,
        response.usage.prompt_tokens,
        response.usage.completion_tokens,
        operation_name,
        server_address,
    )


def assert_all_attributes(
    span: ReadableSpan,
    request_model: str,
    response_id: str = None,
    response_model: str = None,
    input_tokens: Optional[int] = None,
    output_tokens: Optional[int] = None,
    operation_name: str = "chat",
    server_address: str = "api.openai.com",
):
    assert span.name == f"{operation_name} {request_model}"
    assert operation_name == span.attributes[GenAIAttributes.GEN_AI_OPERATION_NAME]
    assert (
        GenAIAttributes.GenAiSystemValues.OPENAI.value
        == span.attributes[GenAIAttributes.GEN_AI_SYSTEM]
    )
    assert request_model == span.attributes[GenAIAttributes.GEN_AI_REQUEST_MODEL]
    if response_model:
        assert response_model == span.attributes[GenAIAttributes.GEN_AI_RESPONSE_MODEL]
    else:
        assert GenAIAttributes.GEN_AI_RESPONSE_MODEL not in span.attributes

    if response_id:
        assert response_id == span.attributes[GenAIAttributes.GEN_AI_RESPONSE_ID]
    else:
        assert GenAIAttributes.GEN_AI_RESPONSE_ID not in span.attributes

    if input_tokens is not None:
        assert (
            input_tokens == span.attributes[GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS]
        )
    else:
        assert GenAIAttributes.GEN_AI_USAGE_INPUT_TOKENS not in span.attributes

    if output_tokens is not None:
        assert (
            output_tokens == span.attributes[GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS]
        )
    else:
        assert GenAIAttributes.GEN_AI_USAGE_OUTPUT_TOKENS not in span.attributes

    assert server_address == span.attributes[ServerAttributes.SERVER_ADDRESS]


def get_current_weather_tool_definition():
    return {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. Boston, MA",
                    },
                },
                "required": ["location"],
                "additionalProperties": False,
            },
        },
    }



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any, Dict, List, Mapping, Optional, Union
from urllib.parse import urlparse

from httpx import URL
from openai import NOT_GIVEN
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion import Choice as OpenAIChoice
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv._incubating.attributes import (
    server_attributes as ServerAttributes,
)

from llm_tracekit import extended_gen_ai_attributes as ExtendedGenAIAttributes
from llm_tracekit.span_builder import (
    Choice,
    Message,
    ToolCall,
    attribute_generator,
    generate_base_attributes,
    generate_choice_attributes,
    generate_message_attributes,
    generate_request_attributes,
    generate_response_attributes,
)


def parse_tool_calls(
    tool_calls: Optional[
        Union[List[Dict[str, Any]], List[ChatCompletionMessageToolCall]]
    ],
) -> Optional[List[ToolCall]]:
    if tool_calls is None:
        return None

    parsed_tool_calls = []

    for tool_call in tool_calls:
        function_name = None
        arguments = None
        func = get_property_value(tool_call, "function")
        if func is not None:
            function_name = get_property_value(func, "name")
            arguments = get_property_value(func, "arguments")
            if isinstance(arguments, str):
                arguments = arguments.replace("\n", "")

        parsed_tool_calls.append(
            ToolCall(
                id=get_property_value(tool_call, "id"),
                type=get_property_value(tool_call, "type"),
                function_name=function_name,
                function_arguments=arguments,
            )
        )

    return parsed_tool_calls


def generate_server_address_and_port_attributes(client_instance) -> Dict[str, Any]:
    base_client = getattr(client_instance, "_client", None)
    base_url = getattr(base_client, "base_url", None)
    if not base_url:
        return {}

    host = None
    port: Optional[int] = -1
    if isinstance(base_url, URL):
        host = base_url.host
        port = base_url.port
    elif isinstance(base_url, str):
        url = urlparse(base_url)
        host = url.hostname
        port = url.port

    attributes: Dict[str, Any] = {ServerAttributes.SERVER_ADDRESS: host}
    if port and port != 443 and port > 0:
        attributes[ServerAttributes.SERVER_PORT] = port

    return attributes


def get_property_value(obj, property_name):
    if isinstance(obj, dict):
        return obj.get(property_name, None)

    return getattr(obj, property_name, None)


def messages_to_span_attributes(
    messages: list, capture_content: bool
) -> Dict[str, Any]:
    parsed_messages = []
    for message in messages:
        content = get_property_value(message, "content")
        if not isinstance(content, str):
            content = None

        tool_calls = parse_tool_calls(get_property_value(message, "tool_calls"))

        parsed_messages.append(
            Message(
                role=get_property_value(message, "role"),
                content=content,
                tool_call_id=get_property_value(message, "tool_call_id"),
                tool_calls=tool_calls,
            )
        )

    return generate_message_attributes(
        messages=parsed_messages, capture_content=capture_content
    )


def choices_to_span_attributes(
    choices: List[OpenAIChoice], capture_content
) -> Dict[str, Any]:
    parsed_choices = []
    for choice in choices:
        role = None
        content = None
        tool_calls = None
        if choice.message:
            role = choice.message.role
            content = choice.message.content
            tool_calls = parse_tool_calls(choice.message.tool_calls)

        parsed_choices.append(
            Choice(
                finish_reason=choice.finish_reason or "error",
                role=role,
                content=content,
                tool_calls=tool_calls,
            )
        )

    return generate_choice_attributes(choices=parsed_choices, capture_content=capture_content)


def set_span_attributes(span, attributes: dict):
    for field, value in attributes.items():
        set_span_attribute(span, field, value)


def set_span_attribute(span, name, value):
    if non_numerical_value_is_set(value) is False:
        return

    span.set_attribute(name, value)


def is_streaming(kwargs):
    return non_numerical_value_is_set(kwargs.get("stream"))


def non_numerical_value_is_set(value: Optional[Union[bool, str]]):
    return bool(value) and value != NOT_GIVEN


@attribute_generator
def get_llm_request_attributes(kwargs, client_instance, capture_content: bool):
    attributes = {
        **generate_base_attributes(system=GenAIAttributes.GenAiSystemValues.OPENAI),
        **generate_request_attributes(
            model=kwargs.get("model"),
            temperature=kwargs.get("temperature"),
            top_p=kwargs.get("p") or kwargs.get("top_p"),
            max_tokens=kwargs.get("max_tokens"),
            presence_penalty=kwargs.get("presence_penalty"),
            frequency_penalty=kwargs.get("frequency_penalty"),
        ),
        **messages_to_span_attributes(
            messages=kwargs.get("messages", []), capture_content=capture_content
        ),
        GenAIAttributes.GEN_AI_OPENAI_REQUEST_SEED: kwargs.get("seed"),
        ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_USER: kwargs.get("user"),
    }

    response_format = kwargs.get("response_format")
    if response_format is not None:
        # response_format may be string or object with a string in the `type` key
        if isinstance(response_format, Mapping):
            response_format_type = response_format.get("type")
            if response_format_type is not None:
                attributes[GenAIAttributes.GEN_AI_OPENAI_REQUEST_RESPONSE_FORMAT] = (
                    response_format_type
                )
        else:
            attributes[GenAIAttributes.GEN_AI_OPENAI_REQUEST_RESPONSE_FORMAT] = (
                response_format
            )

    tools = kwargs.get("tools")
    if tools is not None and isinstance(tools, list):
        for index, tool in enumerate(tools):
            if not isinstance(tool, Mapping):
                continue

            attributes[
                ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_TOOLS_TYPE.format(
                    tool_index=index
                )
            ] = tool.get("type", "function")
            function = tool.get("function")
            if function is not None and isinstance(function, Mapping):
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_TOOLS_FUNCTION_NAME.format(
                        tool_index=index
                    )
                ] = function.get("name")
                attributes[
                    ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_TOOLS_FUNCTION_DESCRIPTION.format(
                        tool_index=index
                    )
                ] = function.get("description")
                function_parameters = function.get("parameters")
                if function_parameters is not None:
                    attributes[
                        ExtendedGenAIAttributes.GEN_AI_OPENAI_REQUEST_TOOLS_FUNCTION_PARAMETERS.format(
                            tool_index=index
                        )
                    ] = json.dumps(function_parameters)

    attributes.update(generate_server_address_and_port_attributes(client_instance))
    service_tier = kwargs.get("service_tier")
    if service_tier != "auto":
        attributes[GenAIAttributes.GEN_AI_OPENAI_RESPONSE_SERVICE_TIER] = service_tier

    return attributes


@attribute_generator
def get_llm_response_attributes(
    result: ChatCompletion, capture_content: bool
) -> Dict[str, Any]:
    finish_reasons = None
    if result.choices is not None:
        finish_reasons = []
        for choice in result.choices:
            finish_reasons.append(choice.finish_reason or "error")

    usage_input_tokens = None
    usage_output_tokens = None
    if result.usage is not None:
        usage_input_tokens = result.usage.prompt_tokens
        usage_output_tokens = result.usage.completion_tokens

    return {
        GenAIAttributes.GEN_AI_OPENAI_REQUEST_SERVICE_TIER: result.service_tier,
        **generate_response_attributes(
            model=result.model,
            finish_reasons=finish_reasons,
            id=result.id,
            usage_input_tokens=usage_input_tokens,
            usage_output_tokens=usage_output_tokens,
        ),
        **choices_to_span_attributes(result.choices, capture_content),
    }
